/*
    diffn - compare a set of files for equality


    This program, diffn, is Copyright (c) 2008-2010 Volker Schatz.
  
    diffn is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 3 of the License, or
    (at your option) any later version.
  
    diffn is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
  
    You should have received a copy of the GNU General Public License
    along with diffn; if not, see <http://www.gnu.org/licenses/gpl.html>.


    compile with:  grep ^gcc diffn.c | sh
gcc -o diffn diffn.c
*/

#define _XOPEN_SOURCE   500         /*  for lstat()  */

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

#ifndef CHUNKSIZE
#define CHUNKSIZE               (1 << 16)
#endif

typedef struct fi {
    char *fname;
    struct stat st;
    int symlink;
    struct fi *link;
    FILE *hdl;
    char *buf;
    int size, last, done, single;
}
fileinfo;

typedef struct {
    char *shortopt, *longopt;
    void *value;
}
cmdswitch;

#define UNCHANGED   "\001\002\003\004"
#define DEFEQSEP    "\n == "
#define DEFIDSEP    "\n   === "

struct {
    int showhelp;               /*  -h, --help  */
    int uselstat;               /*  -l,  --nofollow  */
    int quietall, quietany;     /*  -q, --quiet-all; -Q, --quiet-any  */
    int quiet;                  /*  quietall or quietany  */
    char *neqsep, *eqsep, *idsep;
}
opts= {
    0, 0, 0, 0, 0,
    "\n", UNCHANGED, UNCHANGED
};

cmdswitch cmdlineswitches[] = {
    "-h", "--help", &opts.showhelp,
    "-l", "--nofollow", &opts.uselstat,
    "-q", "--quiet-all", &opts.quietall,
    "-Q", "--quiet-any", &opts.quietany,
    "", "--neqsep", &opts.neqsep,
    "", "--eqsep", &opts.eqsep,
    "", "--idsep", &opts.idsep,
    NULL };

void printwithlinks(const fileinfo *f);
void usageexit(int status);

int main(int argc, char **argv)
{
    cmdswitch *sw;
    fileinfo exch, *files;
    char *neqtab;
    int argind, nfiles, statret, ind, ind2, done, firstprint;

    /*  parse command-line options  */
    for( argind= 1; argind< argc; ++argind ) {
        if( argv[argind][0] != '-' )
            break;
        if( !strcmp(argv[argind], "--") ) {
            ++argind;
            break;
        }
        for( sw= cmdlineswitches; sw->shortopt; ++sw )
            if( !strcmp(argv[argind], sw->shortopt) ||
                !strcmp(argv[argind], sw->longopt) )
                break;
        if( !sw->shortopt ) {
            fprintf(stderr, "Unknown command-line switch `%s'.\n", argv[argind]);
            usageexit(2);
        }
        if( *sw->shortopt )     /*  these are all boolean options  */
            *(int*)sw->value= 1;
        else {                  /*  separator string options  */
            if( argind + 1 >= argc ) {
                fprintf(stderr, "The option `%s' requires an argument.\n", argv[argind]);
                usageexit(2);
            }
            ++argind;
            *(char**)sw->value= argv[argind];
        }
    }
    if( opts.showhelp )
        usageexit(0);
    if( !strcmp(opts.eqsep, UNCHANGED) ) {
        opts.eqsep= DEFEQSEP;
        if( !strcmp(opts.idsep, UNCHANGED) )
            opts.idsep= DEFIDSEP;
    }
    else if( !strcmp(opts.idsep, UNCHANGED) ) 
        opts.idsep= opts.eqsep;
    opts.quiet= opts.quietall || opts.quietany;
    nfiles= argc - argind;
    if( !nfiles )
        return 0;
    neqtab= malloc(nfiles*nfiles);
    files= malloc(sizeof(fileinfo)*nfiles);
    if( !neqtab || !files ) {
        perror("Could not allocate storage for file properties");
        return 2;
    }
    memset(neqtab, 0, nfiles*nfiles);
    /*  initialise file info and stat files  */
    for( ind= 0; ind< nfiles; ++ind ) {
        files[ind].fname= argv[ind+argind];
        if( opts.uselstat ) {
            statret= lstat(files[ind].fname, &files[ind].st);
            if( statret ) {
                perror("Error stat'ing file");
                fprintf(stderr, "Could not stat %s.\n", files[ind].fname);
                return 3;
            }
            files[ind].symlink= S_ISLNK(files[ind].st.st_mode);
        }
        else
            /*  compare target not symlink  */
            files[ind].symlink= 0;
        statret= stat(files[ind].fname, &files[ind].st);
        if( statret ) {
            perror("Error stat'ing file");
            if( !opts.uselstat )
                files[ind].symlink= !lstat(files[ind].fname, &files[ind].st)
                                    && S_ISLNK(files[ind].st.st_mode);
            /*  Complain about target if symlink exists.  */
            fprintf(stderr, "Could not stat %s%s.\n", files[ind].symlink?
                            "target of symlink " : "", files[ind].fname);
            return 3;
        }
        files[ind].link= NULL;
        files[ind].done= 0;
    }
    /*  shortcut 1: file names referring to same inode => equal  */
    for( ind= 0; ind< nfiles; ++ind ) {
        for( ind2= ind+1; ind2< nfiles; ++ind2 )
            if( files[ind].st.st_ino == files[ind2].st.st_ino &&
                    files[ind].symlink == files[ind2].symlink ) {
                if( opts.quietall )
                    return 0;
                if( ind2 < nfiles-1 ) {
                    exch= files[nfiles-1];
                    files[nfiles-1]= files[ind2];
                    files[ind2]= exch;
                }
                files[nfiles-1].link= files[ind].link;
                files[ind].link= files + nfiles - 1;
                --nfiles;
            }
    }
    firstprint= 1;
    /*  shortcut 2: files with different type or size or device number => not
    equal.  Also, directories, sockets and fifos are not equal unless they are
    identical (same inode, see above).  */
    for( ind= 0; ind< nfiles; ++ind )
        files[ind].single= 1;
    done= 1;
    for( ind= 0; ind< nfiles; ++ind ) {
        if( files[ind].done )
            continue;
        for( ind2= ind+1; ind2< nfiles; ++ind2 )
            if( (files[ind].st.st_mode & S_IFMT) != (files[ind2].st.st_mode & S_IFMT)
                    || files[ind].symlink != files[ind2].symlink
                    || S_ISDIR(files[ind].st.st_mode)
                    || S_ISFIFO(files[ind].st.st_mode)
                    || S_ISSOCK(files[ind].st.st_mode)
                    || (S_ISREG(files[ind].st.st_mode) &&
                        files[ind].st.st_size != files[ind2].st.st_size)
                    || ((S_ISBLK(files[ind].st.st_mode) ||
                         S_ISCHR(files[ind].st.st_mode & S_IFMT)) &&
                          files[ind].st.st_dev != files[ind2].st.st_dev)
                ) {
                neqtab[nfiles*ind+ind2]= 1;
                if( opts.quietany )
                    return 1;
            }
            else {
                files[ind].single= 0;
                files[ind2].single= 0;
            }
        if( files[ind].single ) {
            if( !opts.quiet ) {
                if( firstprint )     firstprint= 0;
                else            printf("%s", opts.neqsep);
                printwithlinks(files+ind);
            }
            files[ind].done= 1;
        }
        else if( !S_ISREG(files[ind].st.st_mode) ) {
            if( !opts.quiet ) {
                if( firstprint )     firstprint= 0;
                else            printf("%s", opts.neqsep);
                printwithlinks(files+ind);
            }
            files[ind].done= 1;
            for( ind2= ind+1; ind2< nfiles; ++ind2 )
                if( !neqtab[nfiles*ind+ind2] ) {
                    if( !opts.quiet ) {
                        printf("%s", opts.eqsep);
                        printwithlinks(files+ind2);
                    }
                    files[ind2].done= 1;
                }
        }
        else
            done= 0;
    }
    if( done )
        goto alldone;
    files[0].buf= malloc(CHUNKSIZE*nfiles);
    if( !files[0].buf ) {
        perror("Could not allocate file data buffers");
        return 2;
    }
    /*  initialise buffer pointers and open files  */
    for( ind= 0; ind< nfiles; ++ind ) {
        files[ind].hdl= fopen(files[ind].fname, "rb");
        if( !files[ind].hdl ) {
            perror("Error opening file");
            fprintf(stderr, "Could not open %s for reading.\n", files[ind].fname);
            return 4;
        }
        if( ind )
            files[ind].buf= files[0].buf + ind*CHUNKSIZE;
    }
    /*  compare file contents chunk for chunk  */
    do {
        done= 1;
        /*  read next chunk  */
        for( ind= 0; ind< nfiles; ++ind ) {
            files[ind].single= 1;
            if( !files[ind].done ) {
                files[ind].size= fread(files[ind].buf, 1, CHUNKSIZE, files[ind].hdl);
                if( files[ind].size != CHUNKSIZE && ferror(files[ind].hdl) ) {
                    perror("Error reading file");
                    fprintf(stderr, "Error reading %s.\n", files[ind].fname);
                    return 4;
                }
                files[ind].last= feof(files[ind].hdl);
            }
        }
        /*  compare file read status and data  */
        for( ind= 0; ind< nfiles; ++ind ) {
            if( files[ind].done )
                continue;
            for( ind2= ind+1; ind2< nfiles; ++ind2 ) {
                if( neqtab[nfiles*ind+ind2] )
                    continue;
                if( files[ind].size != files[ind2].size ||
                    files[ind].last != files[ind2].last ||
                    memcmp(files[ind].buf, files[ind2].buf, files[ind].size) )
                {
                    neqtab[nfiles*ind+ind2]= 1;
                    if( opts.quietany )
                        return 1;
                }
                else {
                    files[ind].single= 0;
                    files[ind2].single= 0;
                }
            }
            if( files[ind].single ) {
                if( !opts.quiet ) {
                    if( firstprint )     firstprint= 0;
                    else            printf("%s", opts.neqsep);
                    printwithlinks(files+ind);
                }
                files[ind].done= 1;
                fclose(files[ind].hdl);
            }
            else if( files[ind].last ) {
                if( !opts.quiet ) {
                    if( firstprint )     firstprint= 0;
                    else            printf("%s", opts.neqsep);
                    printwithlinks(files+ind);
                }
                files[ind].done= 1;
                fclose(files[ind].hdl);
                for( ind2= ind+1; ind2< nfiles; ++ind2 )
                    if( !neqtab[nfiles*ind+ind2] ) {
                        if( !opts.quiet ) {
                            printf("%s", opts.eqsep);
                            printwithlinks(files+ind2);
                        }
                        files[ind2].done= 1;
                        fclose(files[ind2].hdl);
                    }
            }
            else
                done= 0;
        }
    }
      while( !done );
alldone:
    if( !opts.quiet )
        printf("\n");
    if( opts.quietall ) {
        for( ind= 0; ind< nfiles; ++ind ) {
        /*  if( files[ind].link )
                goto notalldiff;        - already handled at shortcut 1  */
            for( ind2= ind+1; ind2< nfiles; ++ind2 )
                if( !neqtab[nfiles*ind+ind2] )
                    goto notalldiff;
        }
        return 1;
notalldiff: ;
    }
    return 0;
}


/*
    Print file name and all its hard links.
*/
void printwithlinks(const fileinfo *f)
{
    fileinfo *ln;

    printf("%s", f->fname);
    for( f= f->link; f; f= f->link )
        printf("%s%s", opts.idsep, f->fname);
}


/*
    Print usage message and exit/
*/
void usageexit(int status)
{
    printf("\ndiffn - compare a set of files for equality\n\n"
"usage: diffn [<options>] files...\n\n"
"Outputs which of the given files are identical (refer to the same inode) or\n"
"equal (have the same content, or are equivalent special files).\n\n"
"Command-line options:\n"
"-h, --help       show this message\n"
"-l, --nofollow   do not dereference symbolic links\n"
"-q, --quiet-all  do not output anything, but return 1 if all files are unique\n"
"-Q, --quiet-any  no output, but return 1 if not all files are the same\n"
"--neqsep, --eqsep, --idsep set separator strings between file names which were\n"
"found to be unequal, equal or identical\n\n"
"If diffn is correctly installed on your system, the command `man diffn' should\n"
"display the full manual.\n");
    exit(status);
}

